import path from 'node:path';
import mapWorkspaces from '@npmcli/map-workspaces';
import micromatch from 'micromatch';
import { ConfigurationValidator } from './configuration-validator.js';
import { ROOT_WORKSPACE_NAME, DEFAULT_WORKSPACE_CONFIG, KNIP_CONFIG_LOCATIONS } from './constants.js';
import * as plugins from './plugins/index.js';
import { arrayify } from './util/array.js';
import parsedArgs from './util/cli-arguments.js';
import { ConfigurationError } from './util/errors.js';
import { findFile, loadJSON } from './util/fs.js';
import { ensurePosixPath } from './util/glob.js';
import { _load } from './util/loader.js';
import { toCamelCase } from './util/plugin.js';
import { resolveIncludedIssueTypes } from './util/resolve-included-issue-types.js';
import { byPathDepth } from './util/workspace.js';
const { values: { config: rawConfigArg, workspace: rawWorkspaceArg, include = [], exclude = [], dependencies = false, exports = false, }, } = parsedArgs;
const defaultWorkspaceConfig = DEFAULT_WORKSPACE_CONFIG;
const defaultConfig = {
    include: [],
    exclude: [],
    ignore: [],
    ignoreBinaries: [],
    ignoreDependencies: [],
    ignoreWorkspaces: [],
    workspaces: {
        [ROOT_WORKSPACE_NAME]: defaultWorkspaceConfig,
    },
};
const PLUGIN_NAMES = Object.keys(plugins);
export default class ConfigurationChief {
    cwd;
    isProduction = false;
    config;
    manifestPath;
    manifest;
    manifestWorkspaces;
    resolvedConfigFilePath;
    constructor({ cwd, isProduction }) {
        this.cwd = cwd;
        this.isProduction = isProduction;
        this.config = defaultConfig;
    }
    async loadLocalConfig() {
        const manifestPath = findFile(this.cwd, 'package.json');
        const manifest = manifestPath && (await loadJSON(manifestPath));
        if (!manifestPath || !manifest) {
            throw new ConfigurationError('Unable to find package.json');
        }
        this.manifestPath = manifestPath;
        this.manifest = manifest;
        const pnpmWorkspacesPath = findFile(this.cwd, 'pnpm-workspace.yaml');
        const pnpmWorkspaces = pnpmWorkspacesPath && (await _load(pnpmWorkspacesPath));
        if (this.manifest && !this.manifest.workspaces && pnpmWorkspaces) {
            this.manifest.workspaces = pnpmWorkspaces;
        }
        for (const configPath of rawConfigArg ? [rawConfigArg] : KNIP_CONFIG_LOCATIONS) {
            this.resolvedConfigFilePath = findFile(this.cwd, configPath);
            if (this.resolvedConfigFilePath)
                break;
        }
        if (rawConfigArg && !this.resolvedConfigFilePath && !manifest.knip) {
            throw new ConfigurationError(`Unable to find ${rawConfigArg} or package.json#knip`);
        }
        const rawLocalConfig = this.resolvedConfigFilePath ? await _load(this.resolvedConfigFilePath) : manifest.knip;
        if (rawLocalConfig) {
            this.config = this.normalize(ConfigurationValidator.parse(rawLocalConfig));
        }
    }
    normalize(rawLocalConfig) {
        const workspaces = rawLocalConfig.workspaces ?? {
            [ROOT_WORKSPACE_NAME]: {
                ...rawLocalConfig,
            },
        };
        const include = rawLocalConfig.include ?? defaultConfig.include;
        const exclude = rawLocalConfig.exclude ?? defaultConfig.exclude;
        const ignore = arrayify(rawLocalConfig.ignore ?? defaultConfig.ignore);
        const ignoreBinaries = rawLocalConfig.ignoreBinaries ?? defaultConfig.ignoreBinaries;
        const ignoreDependencies = rawLocalConfig.ignoreDependencies ?? defaultConfig.ignoreDependencies;
        const ignoreWorkspaces = rawLocalConfig.ignoreWorkspaces ?? defaultConfig.ignoreWorkspaces;
        return {
            include,
            exclude,
            ignore,
            ignoreBinaries,
            ignoreDependencies,
            ignoreWorkspaces,
            workspaces: Object.entries(workspaces)
                .filter(([workspaceName]) => !ignoreWorkspaces.includes(workspaceName))
                .reduce((workspaces, workspace) => {
                const [workspaceName, workspaceConfig] = workspace;
                const entry = workspaceConfig.entry ? arrayify(workspaceConfig.entry) : defaultWorkspaceConfig.entry;
                const project = workspaceConfig.project
                    ? arrayify(workspaceConfig.project)
                    : workspaceConfig.entry
                        ? entry
                        : defaultWorkspaceConfig.project;
                const paths = workspaceConfig.paths ?? defaultWorkspaceConfig.paths;
                workspaces[workspaceName] = {
                    entry,
                    project,
                    paths,
                    ignore: arrayify(workspaceConfig.ignore),
                };
                for (const [name, pluginConfig] of Object.entries(workspaceConfig)) {
                    const pluginName = toCamelCase(name);
                    if (PLUGIN_NAMES.includes(pluginName)) {
                        if (pluginConfig === false) {
                            workspaces[workspaceName][pluginName] = false;
                        }
                        else {
                            const isObject = typeof pluginConfig !== 'string' && !Array.isArray(pluginConfig);
                            const config = isObject ? arrayify(pluginConfig.config) : pluginConfig ? arrayify(pluginConfig) : null;
                            const entry = isObject && 'entry' in pluginConfig ? arrayify(pluginConfig.entry) : null;
                            const project = isObject && 'project' in pluginConfig ? arrayify(pluginConfig.project) : entry;
                            workspaces[workspaceName][pluginName] = {
                                config,
                                entry,
                                project,
                            };
                        }
                    }
                }
                return workspaces;
            }, {}),
        };
    }
    async getManifestWorkspaces() {
        if (this.manifestWorkspaces)
            return this.manifestWorkspaces;
        if (this.manifest) {
            const workspaces = await mapWorkspaces({
                pkg: this.manifest,
                cwd: this.cwd,
                ignore: this.config.ignoreWorkspaces,
                absolute: false,
            });
            this.manifestWorkspaces = Array.from(workspaces.values())
                .map(dir => path.relative(this.cwd, dir))
                .map(ensurePosixPath);
            return this.manifestWorkspaces;
        }
        return [];
    }
    getConfiguredWorkspaces() {
        return this.config.workspaces ? Object.keys(this.config.workspaces) : [];
    }
    getAdditionalWorkspaces(manifestWorkspaces) {
        return Object.keys(this.config.workspaces).filter(name => !name.includes('*') && !manifestWorkspaces.includes(name));
    }
    async getActiveWorkspaces() {
        const manifestWorkspaces = await this.getManifestWorkspaces();
        const additionalWorkspaces = this.getAdditionalWorkspaces(manifestWorkspaces);
        const rootWorkspace = {
            name: ROOT_WORKSPACE_NAME,
            dir: this.cwd,
            config: this.getConfigForWorkspace(ROOT_WORKSPACE_NAME),
            ancestors: [],
        };
        const isOnlyRootWorkspace = (manifestWorkspaces.length === 0 && !rawWorkspaceArg) ||
            (rawWorkspaceArg && [ROOT_WORKSPACE_NAME, './'].includes(rawWorkspaceArg));
        if (isOnlyRootWorkspace)
            return [rootWorkspace];
        if (rawWorkspaceArg) {
            const workspace = {
                name: rawWorkspaceArg,
                dir: path.resolve(this.cwd, rawWorkspaceArg),
                config: this.getConfigForWorkspace(rawWorkspaceArg),
                ancestors: [ROOT_WORKSPACE_NAME],
            };
            return this.hasConfigForWorkspace(ROOT_WORKSPACE_NAME) ? [rootWorkspace, workspace] : [workspace];
        }
        const workspaces = [...manifestWorkspaces, ...additionalWorkspaces];
        const activeWorkspaces = workspaces.filter(workspaceName => this.hasConfigForWorkspace(workspaceName));
        return activeWorkspaces.sort(byPathDepth).map(name => ({
            name,
            dir: path.resolve(this.cwd, name),
            config: this.getConfigForWorkspace(name),
            ancestors: activeWorkspaces.reduce((ancestors, ancestorName) => {
                if (name === ancestorName)
                    return ancestors;
                if (ancestorName === ROOT_WORKSPACE_NAME || name.startsWith(ancestorName)) {
                    if (this.hasConfigForWorkspace(ancestorName)) {
                        ancestors.push(ancestorName);
                    }
                }
                return ancestors;
            }, []),
        }));
    }
    async getDescendentWorkspaces(name) {
        const manifestWorkspaces = await this.getManifestWorkspaces();
        const additionalWorkspaces = this.getAdditionalWorkspaces(manifestWorkspaces);
        return [...manifestWorkspaces, ...additionalWorkspaces]
            .filter(workspaceName => workspaceName !== name)
            .filter(workspaceName => name === ROOT_WORKSPACE_NAME || workspaceName.startsWith(name));
    }
    async getNegatedWorkspacePatterns(name) {
        const descendentWorkspaces = await this.getDescendentWorkspaces(name);
        const matchName = new RegExp(`^${name}/`);
        return descendentWorkspaces
            .map(workspaceName => workspaceName.replace(matchName, ''))
            .map(workspaceName => `!${workspaceName}`);
    }
    getConfigKeyForWorkspace(workspaceName) {
        const configuredWorkspaces = this.getConfiguredWorkspaces();
        return configuredWorkspaces
            .sort(byPathDepth)
            .reverse()
            .find(pattern => micromatch.isMatch(workspaceName, pattern));
    }
    hasConfigForWorkspace(workspaceName) {
        return Boolean(this.getConfigKeyForWorkspace(workspaceName));
    }
    getConfigForWorkspace(workspaceName) {
        const key = this.getConfigKeyForWorkspace(workspaceName);
        if (key && this.config?.workspaces?.[key])
            return this.config.workspaces[key];
        return { entry: [], project: [], paths: {}, ignore: [] };
    }
    resolveIncludedIssueTypes() {
        return resolveIncludedIssueTypes({
            include,
            exclude,
            dependencies,
            exports,
        }, {
            include: this.config.include ?? [],
            exclude: this.config.exclude ?? [],
            isProduction: this.isProduction,
        });
    }
}
