import { isBuiltin } from 'node:module';
import micromatch from 'micromatch';
import { IGNORE_DEFINITELY_TYPED } from './constants.js';
import { getPackageNameFromModuleSpecifier, isDefinitelyTyped, getDefinitelyTypedFor, getPackageFromDefinitelyTyped, } from './util/modules.js';
export default class DependencyDeputy {
    isStrict;
    _manifests = new Map();
    manifests = new Map();
    ignoreDependencies;
    referencedDependencies;
    peerDependencies;
    installedBinaries;
    tsConfigPathGlobs = new Map();
    constructor({ isStrict, ignoreDependencies }) {
        this.isStrict = isStrict;
        this.ignoreDependencies = ignoreDependencies;
        this.referencedDependencies = new Map();
        this.peerDependencies = new Map();
        this.installedBinaries = new Map();
    }
    addWorkspace({ name, dir, manifestPath, manifest, }) {
        const scripts = Object.values(manifest.scripts ?? {});
        const dependencies = Object.keys(manifest.dependencies ?? {});
        const peerDependencies = Object.keys(manifest.peerDependencies ?? {});
        const optionalDependencies = Object.keys(manifest.optionalDependencies ?? {});
        const devDependencies = Object.keys(manifest.devDependencies ?? {});
        const allDependencies = [...dependencies, ...devDependencies, ...peerDependencies, ...optionalDependencies];
        this.manifests.set(name, manifest);
        this._manifests.set(name, {
            workspaceDir: dir,
            manifestPath,
            scripts,
            dependencies,
            peerDependencies,
            optionalDependencies,
            devDependencies,
            allDependencies,
        });
    }
    getManifest(workspaceName) {
        return this.manifests.get(workspaceName);
    }
    getWorkspaceManifest(workspaceName) {
        return this._manifests.get(workspaceName);
    }
    getProductionDependencies(workspaceName) {
        return this._manifests.get(workspaceName)?.dependencies ?? [];
    }
    getDevDependencies(workspaceName) {
        return this._manifests.get(workspaceName)?.devDependencies ?? [];
    }
    setInstalledBinaries(workspaceName, installedBinaries) {
        this.installedBinaries.set(workspaceName, installedBinaries);
    }
    getInstalledBinaries(workspaceName) {
        return this.installedBinaries.get(workspaceName);
    }
    addTypeScriptConfigPathGlobs(workspaceName, paths) {
        const wsPaths = this.tsConfigPathGlobs.get(workspaceName) ?? [];
        const addPaths = Object.keys(paths).map(value => value.replace(/\*/g, '**'));
        this.tsConfigPathGlobs.set(workspaceName, [...wsPaths, ...addPaths]);
    }
    addReferencedDependency(workspaceName, packageName) {
        if (!this.referencedDependencies.has(workspaceName)) {
            this.referencedDependencies.set(workspaceName, new Set());
        }
        this.referencedDependencies.get(workspaceName)?.add(packageName);
    }
    addPeerDependencies(workspaceName, peerDependencies) {
        this.peerDependencies.set(workspaceName, peerDependencies);
    }
    getPeerDependencies(workspaceName, dependency) {
        return Array.from(this.peerDependencies.get(workspaceName)?.get(dependency) ?? []);
    }
    maybeAddListedReferencedDependency(workspace, moduleSpecifier) {
        if (this.isInternalDependency(workspace.name, moduleSpecifier))
            return;
        const packageName = getPackageNameFromModuleSpecifier(moduleSpecifier);
        const workspaceNames = this.isStrict ? [workspace.name] : [workspace.name, ...[...workspace.ancestors].reverse()];
        const closestWorkspaceName = workspaceNames.find(name => this.isInDependencies(name, packageName));
        const typesPackageName = !isDefinitelyTyped(packageName) && getDefinitelyTypedFor(packageName);
        const closestWorkspaceNameForTypes = typesPackageName && workspaceNames.find(name => this.isInDependencies(name, typesPackageName));
        if (closestWorkspaceName || closestWorkspaceNameForTypes) {
            closestWorkspaceName && this.addReferencedDependency(closestWorkspaceName, packageName);
            closestWorkspaceNameForTypes && this.addReferencedDependency(closestWorkspaceNameForTypes, typesPackageName);
            return;
        }
        for (const name of workspaceNames) {
            const binaries = this.getInstalledBinaries(name);
            if (binaries?.has(moduleSpecifier)) {
                const dependencies = binaries.get(moduleSpecifier);
                if (dependencies?.size) {
                    dependencies.forEach(dependency => this.addReferencedDependency(name, dependency));
                    return;
                }
            }
        }
        return moduleSpecifier;
    }
    isInternalDependency(workspaceName, moduleSpecifier) {
        if (moduleSpecifier.startsWith('/') || moduleSpecifier.startsWith('.'))
            return true;
        if (isBuiltin(moduleSpecifier))
            return true;
        const packageName = getPackageNameFromModuleSpecifier(moduleSpecifier);
        return (!this.isInDependencies(workspaceName, packageName) && this.isAliasedDependency(workspaceName, moduleSpecifier));
    }
    isAliasedDependency(workspaceName, moduleSpecifier) {
        const patterns = this.tsConfigPathGlobs.get(workspaceName);
        return patterns && patterns.length > 0 && micromatch.isMatch(moduleSpecifier, patterns);
    }
    isInDependencies(workspaceName, packageName) {
        const manifest = this._manifests.get(workspaceName);
        const dependencies = manifest ? (this.isStrict ? manifest.dependencies : manifest.allDependencies) : [];
        return dependencies.includes(packageName);
    }
    settleDependencyIssues() {
        const dependencyIssues = [];
        const devDependencyIssues = [];
        for (const [workspaceName, { manifestPath }] of this._manifests.entries()) {
            const referencedDependencies = this.referencedDependencies.get(workspaceName);
            const isUnreferencedDependency = (dependency) => {
                if (referencedDependencies?.has(dependency))
                    return false;
                const [scope, typedDependency] = dependency.split('/');
                if (scope === '@types') {
                    const typedPackageName = getPackageFromDefinitelyTyped(typedDependency);
                    if (IGNORE_DEFINITELY_TYPED.includes(typedPackageName))
                        return false;
                    const peerDependencies = this.getPeerDependencies(workspaceName, typedPackageName);
                    if (peerDependencies.length) {
                        return !peerDependencies.find(peerDependency => !isUnreferencedDependency(peerDependency));
                    }
                    return !referencedDependencies?.has(typedPackageName);
                }
                if (!referencedDependencies?.has(dependency)) {
                    const peerDependencies = this.getPeerDependencies(workspaceName, dependency);
                    return !peerDependencies.find(peerDependency => !isUnreferencedDependency(peerDependency));
                }
                return false;
            };
            this.getProductionDependencies(workspaceName)
                .filter(symbol => symbol !== 'knip' && !this.ignoreDependencies.includes(symbol))
                .filter(isUnreferencedDependency)
                .forEach(symbol => dependencyIssues.push({ type: 'dependencies', filePath: manifestPath, symbol }));
            this.getDevDependencies(workspaceName)
                .filter(symbol => symbol !== 'knip' && !this.ignoreDependencies.includes(symbol))
                .filter(isUnreferencedDependency)
                .forEach(symbol => devDependencyIssues.push({ type: 'devDependencies', filePath: manifestPath, symbol }));
        }
        return { dependencyIssues, devDependencyIssues };
    }
}
