import { ESLint } from 'eslint';
import { compact } from '../../util/array.js';
import { debugLogFiles, debugLogObject } from '../../util/debug.js';
import { _firstGlob } from '../../util/glob.js';
import { _load } from '../../util/loader.js';
import { getPackageName } from '../../util/modules.js';
import { timerify } from '../../util/performance.js';
import { hasDependency } from '../../util/plugin.js';
import { resolvePluginPackageName, customResolvePluginPackageNames, getDependenciesFromSettings } from './helpers.js';
export const NAME = 'ESLint';
export const ENABLERS = ['eslint'];
export const isEnabled = ({ dependencies }) => hasDependency(dependencies, ENABLERS);
export const CONFIG_FILE_PATTERNS = ['.eslintrc', '.eslintrc.{js,json,cjs}', '.eslintrc.{yml,yaml}', 'package.json'];
export const ENTRY_FILE_PATTERNS = ['eslint.config.js'];
const findESLintDependencies = async (configFilePath, { cwd, manifest, workspaceConfig }) => {
    if (configFilePath.endsWith('package.json') && !manifest.eslintConfig)
        return [];
    const config = configFilePath.endsWith('package.json')
        ? manifest.eslintConfig
        : await _load(configFilePath);
    const rootExtends = config?.extends ? [config.extends].flat().map(customResolvePluginPackageNames) : [];
    if (rootExtends.includes('eslint-plugin-prettier'))
        rootExtends.push('eslint-config-prettier');
    const patterns = compact([
        workspaceConfig.entry,
        ...(config?.overrides?.map(overrides => [overrides.files].flat()) ?? []),
    ]);
    debugLogObject('ESLint overrides file patterns', patterns);
    const samples = await Promise.all(patterns.map(patterns => _firstGlob({ patterns, cwd })));
    const sampleFilePaths = compact(samples.filter((filePath) => typeof filePath !== 'undefined').map(String));
    debugLogFiles('ESLint overrides sample files', sampleFilePaths);
    const engine = new ESLint({ cwd, overrideConfigFile: configFilePath, useEslintrc: false });
    const calculateConfigForFile = async (sampleFile) => await engine.calculateConfigForFile(sampleFile);
    const dependencies = await Promise.all(sampleFilePaths.map(calculateConfigForFile)).then(configs => configs.flatMap(config => {
        if (!config)
            return [];
        const plugins = config.plugins?.map(resolvePluginPackageName) ?? [];
        const parsers = config.parser ? [config.parser] : [];
        const extraParsers = config.parserOptions?.babelOptions?.presets ?? [];
        const settings = config.settings ? getDependenciesFromSettings(config.settings) : [];
        return [...parsers, ...extraParsers, ...plugins, ...settings].map(getPackageName);
    }));
    return compact([...rootExtends, ...dependencies.flat()]);
};
export const findDependencies = timerify(findESLintDependencies);
