import path from 'node:path';
import { ts } from 'ts-morph';
import { debugLogSourceFiles } from './util/debug.js';
import { _createProject, partitionSourceFiles, _resolveSourceFileDependencies, _removeExternalSourceFiles, } from './util/project.js';
const baseCompilerOptions = {
    allowJs: true,
    allowSyntheticDefaultImports: true,
    jsx: ts.JsxEmit.ReactJSX,
    esModuleInterop: true,
    skipDefaultLibCheck: true,
    skipLibCheck: true,
    target: ts.ScriptTarget.ES2022,
    module: ts.ModuleKind.CommonJS,
    moduleResolution: ts.ModuleResolutionKind.NodeJs,
};
const skipAddFiles = { skipAddingFilesFromTsConfig: true, skipFileDependencyResolution: true };
export default class ProjectPrincipal {
    projectOptions;
    tsConfigFilePath;
    entryPaths = new Set();
    projectPaths = new Set();
    entryWorkspace;
    projectWorkspace;
    entryFiles;
    addEntryPath(filePath) {
        this.entryPaths.add(filePath);
    }
    addSourceFile(filePath) {
        this.entryWorkspace?.addSourceFileAtPath(filePath);
    }
    addProjectPath(filePath) {
        this.projectPaths.add(filePath);
    }
    removeProjectPath(filePath) {
        this.projectPaths.delete(filePath);
    }
    addTypeScriptPaths(workspaceDir, compilerOptions) {
        if (!compilerOptions || !compilerOptions.paths)
            return;
        if (!this.projectOptions)
            this.projectOptions = {};
        if (!this.projectOptions.compilerOptions)
            this.projectOptions.compilerOptions = {};
        if (!this.projectOptions.compilerOptions.paths)
            this.projectOptions.compilerOptions.paths = {};
        const { baseUrl, paths } = compilerOptions;
        for (const [key, entries] of Object.entries(paths)) {
            const workspacePaths = entries.map(entry => baseUrl ? path.join(workspaceDir, baseUrl, entry) : path.join(workspaceDir, entry));
            if (this.projectOptions.compilerOptions.paths[key]) {
                this.projectOptions.compilerOptions.paths[key].push(...workspacePaths);
            }
            else {
                this.projectOptions.compilerOptions.paths[key] = workspacePaths;
            }
        }
    }
    createProjects() {
        const compilerOptions = { ...baseCompilerOptions, ...this.projectOptions?.compilerOptions };
        this.projectOptions = this.tsConfigFilePath
            ? { tsConfigFilePath: this.tsConfigFilePath, compilerOptions }
            : { compilerOptions };
        this.entryWorkspace = _createProject({ ...this.projectOptions, ...skipAddFiles }, [...this.entryPaths]);
        this.entryFiles = this.entryWorkspace.getSourceFiles();
        debugLogSourceFiles(`Included entry files`, this.entryFiles);
        this.projectWorkspace = _createProject({ ...this.projectOptions, ...skipAddFiles }, [...this.projectPaths]);
        const projectFiles = this.projectWorkspace.getSourceFiles();
        debugLogSourceFiles('Included project files', projectFiles);
    }
    getResolvedFiles() {
        if (!this.entryWorkspace)
            return [];
        _resolveSourceFileDependencies(this.entryWorkspace);
        return _removeExternalSourceFiles(this.entryWorkspace);
    }
    settleFiles() {
        const entryFiles = this.entryFiles;
        const projectFiles = this.projectWorkspace?.getSourceFiles();
        if (this.entryWorkspace && entryFiles && projectFiles) {
            const resolvedFiles = this.getResolvedFiles();
            debugLogSourceFiles(`Included files resolved from entry files`, resolvedFiles);
            const [usedResolvedFiles, unreferencedResolvedFiles] = partitionSourceFiles(projectFiles, resolvedFiles);
            const [usedEntryFiles, usedNonEntryFiles] = partitionSourceFiles(usedResolvedFiles, entryFiles);
            debugLogSourceFiles('Used files', usedResolvedFiles);
            debugLogSourceFiles('Unreferenced files', unreferencedResolvedFiles);
            debugLogSourceFiles('Used entry files', usedEntryFiles);
            debugLogSourceFiles('Used non-entry files', usedNonEntryFiles);
            return { usedResolvedFiles, unreferencedResolvedFiles };
        }
        const emptySet = new Set();
        return { usedResolvedFiles: emptySet, unreferencedResolvedFiles: emptySet };
    }
}
