import { performance, PerformanceObserver } from 'node:perf_hooks';
import EasyTable from 'easy-table';
import prettyMilliseconds from 'pretty-ms';
import parsedArgs from './cli-arguments.js';
import Summary from 'summary';
const { values } = parsedArgs;
const { performance: isEnabled = false } = values;
export const timerify = fn => (isEnabled ? performance.timerify(fn) : fn);
class Performance {
    enabled;
    startTime = 0;
    entries = [];
    instanceId;
    observer;
    constructor(enabled) {
        if (enabled) {
            this.startTime = performance.now();
            this.instanceId = Math.floor(performance.now() * 100);
            this.observer = new PerformanceObserver(items => {
                items.getEntries().forEach(entry => {
                    this.entries.push(entry);
                });
            });
            this.observer.observe({ entryTypes: ['measure', 'function'] });
        }
        this.enabled = enabled;
    }
    start(name) {
        if (!this.enabled)
            return;
        const id = `${this.instanceId}:${name}`;
        performance.mark(`${id}:start`);
    }
    end(name) {
        if (!this.enabled)
            return;
        const id = `${this.instanceId}:${name}`;
        performance.mark(`${id}:end`);
        performance.measure(id, `${id}:start`, `${id}:end`);
        performance.clearMarks(`${id}:start`);
        performance.clearMarks(`${id}:end`);
    }
    getEntriesByName() {
        return this.entries.reduce((entries, entry) => {
            const name = entry.name.replace(`${this.instanceId}:`, '');
            entries[name] = entries[name] ?? [];
            entries[name].push(entry.duration);
            return entries;
        }, {});
    }
    getTable() {
        const entriesByName = this.getEntriesByName();
        const table = new EasyTable();
        Object.entries(entriesByName).map(([name, values]) => {
            const stats = new Summary(values);
            table.cell('Name', name);
            table.cell('size', stats.size(), EasyTable.number(0));
            table.cell('min', stats.min(), EasyTable.number(2));
            table.cell('max', stats.max(), EasyTable.number(2));
            table.cell('median', stats.median(), EasyTable.number(2));
            table.cell('sum', stats.sum(), EasyTable.number(2));
            table.newRow();
        });
        table.sort(['sum|des']);
        return table;
    }
    async flush() {
        this.start('_flush');
        await new Promise(resolve => setTimeout(resolve, 1));
        this.end('_flush');
    }
    async print() {
        if (!this.enabled)
            return;
        await this.flush();
        console.log('\n' + this.getTable().toString().trim());
        console.log('\nTotal running time:', prettyMilliseconds(performance.now() - this.startTime));
    }
    reset() {
        this.entries = [];
        this.observer?.disconnect();
    }
}
export const measure = new Performance(isEnabled);
